test_that("cross-sectional gaussian", {
  
  gdsfile  <- system.file("extdata", "geno.gds",  package = "MAGEE")
  bgenfile <- system.file("extdata", "geno.bgen", package = "MAGEE")
  samplefile <- system.file("extdata", "geno.sample", package = "MAGEE")
  group.file <- system.file("extdata", "SetID.singlevariant.txt", package = "MAGEE")
  data(example)
  set.seed(123)
  pheno <- rbind(example$pheno, example$pheno[1:100, ])
  pheno$id <- 1:500
  pheno$disease[sample(1:500,20)] <- NA
  pheno$age[sample(1:500,20)] <- NA
  pheno$sex[sample(1:500,20)] <- NA
  pheno <- pheno[sample(1:500,450), ]
  pheno <- pheno[pheno$id <= 400, ]
  kins <- example$GRM
  
  ### single thread with kins
  obj1 <- glmmkin(trait ~ age + sex, data = pheno, kins = kins, id = "id", family = gaussian(link = "identity"))
  out1 <- MAGEE(null.obj=obj1, interaction="sex", geno.file=gdsfile, group.file=group.file, tests = c("JV", "JF", "JD"), use.minor.allele = T, auto.flip = T)
  obj1.outfile.gds <- tempfile()
  glmm.gei(null.obj=obj1, interaction="sex", geno.file=gdsfile, outfile=obj1.outfile.gds)
  obj1.gds <- read.table(obj1.outfile.gds, header = T, as.is = T)
  expect_equal(signif(range(obj1.gds$PVAL.MARGINAL)), signif(c(0.007431214, 0.993737081)))
  expect_equal(signif(range(obj1.gds$PVAL.INT)), signif(c(0.02633466, 0.99253758)))
  expect_equal(signif(range(obj1.gds$PVAL.JOINT)), signif(c(0.00626638, 0.98730215)))
  expect_equal(signif(out1$MV.pval,4), signif(obj1.gds$PVAL.MARGINAL,4))
  expect_equal(signif(out1$MF.pval), signif(obj1.gds$PVAL.MARGINAL))
  expect_equal(signif(out1$IV.pval,4), signif(obj1.gds$PVAL.INT,4))
  expect_equal(signif(out1$IF.pval), signif(obj1.gds$PVAL.INT))
  expect_equal(signif(pchisq(qchisq(out1$MV.pval,2,lower=F)+qchisq(out1$IV.pval,2,lower=F),4,lower=F)), signif(out1$JV.pval))
  expect_equal(signif(out1$JV.pval,4), signif(out1$JF.pval,4))
  expect_equal(signif(out1$JV.pval,4), signif(out1$JD.pval,4))
  expect_equal(signif(pchisq(qchisq(obj1.gds$PVAL.MARGINAL,1,lower=F)+qchisq(obj1.gds$PVAL.INT,1,lower=F),2,lower=F)), signif(obj1.gds$PVAL.JOINT))

  obj1.outfile.bgen <- tempfile()
  glmm.gei(null.obj=obj1, interaction="sex", geno.file=bgenfile, bgen.samplefile = samplefile, outfile=obj1.outfile.bgen)
  obj1.bgen <- read.table(obj1.outfile.bgen, header = T, as.is = T)
#  expect_equal(obj1.gds[,-(2:3)], obj1.bgen[,-2])
  unlink(c(obj1.outfile.gds, obj1.outfile.bgen))

  obj1.glmm.score.outfile.gds <- tempfile()
  glmm.score(obj=obj1, infile=gdsfile, outfile=obj1.glmm.score.outfile.gds)
  obj1.glmm.score.gds <- read.table(obj1.glmm.score.outfile.gds,header=T,as.is=T)
  expect_equal(obj1.gds$PVAL.MARGINAL, obj1.glmm.score.gds$PVAL)
  obj1.glmm.score.outfile.bgen <- tempfile()
  glmm.score(obj=obj1, infile=bgenfile, BGEN.samplefile = samplefile, outfile=obj1.glmm.score.outfile.bgen)
  obj1.glmm.score.bgen <- read.table(obj1.glmm.score.outfile.bgen,header=T,as.is=T)
  expect_equal(signif(obj1.gds$PVAL.MARGINAL), signif(obj1.glmm.score.bgen$PVAL))
  unlink(c(obj1.glmm.score.outfile.gds, obj1.glmm.score.outfile.bgen))

  ### single thread without kins
  obj2 <- glmmkin(trait ~ age + sex, data = pheno, kins = NULL, id = "id", family = gaussian(link = "identity"))
  out2 <- MAGEE(null.obj=obj2, interaction="sex", geno.file=gdsfile, group.file=group.file, tests = c("JV", "JF", "JD"), use.minor.allele = T, auto.flip = T)
  obj2.outfile.gds <- tempfile()
  glmm.gei(null.obj=obj2, interaction="sex", geno.file=gdsfile, outfile=obj2.outfile.gds)
  obj2.gds <- read.table(obj2.outfile.gds, header = T, as.is = T)
  expect_equal(signif(range(obj2.gds$PVAL.MARGINAL)), signif(c(0.0002174546, 0.9996967239)))
  expect_equal(signif(range(obj2.gds$PVAL.INT)), signif(c(0.02536793, 0.99887262)))
  expect_equal(signif(range(obj2.gds$PVAL.JOINT)), signif(c(0.0006177245, 0.9979718241)))
  expect_equal(signif(out2$MV.pval,3), signif(obj2.gds$PVAL.MARGINAL,3))
  expect_equal(signif(out2$MF.pval), signif(obj2.gds$PVAL.MARGINAL))
  expect_equal(signif(out2$IV.pval,3), signif(obj2.gds$PVAL.INT,3))
  expect_equal(signif(out2$IF.pval), signif(obj2.gds$PVAL.INT))
  expect_equal(signif(pchisq(qchisq(out2$MV.pval,2,lower=F)+qchisq(out2$IV.pval,2,lower=F),4,lower=F)), signif(out2$JV.pval))
  expect_equal(signif(out2$JV.pval,5), signif(out2$JF.pval,5))
  expect_equal(signif(out2$JV.pval,5), signif(out2$JD.pval,5))
  expect_equal(signif(pchisq(qchisq(obj2.gds$PVAL.MARGINAL,1,lower=F)+qchisq(obj2.gds$PVAL.INT,1,lower=F),2,lower=F)), signif(obj2.gds$PVAL.JOINT))

  obj2.outfile.bgen <- tempfile()
  glmm.gei(null.obj=obj2, interaction="sex", geno.file=bgenfile, bgen.samplefile = samplefile, outfile=obj2.outfile.bgen)
  obj2.bgen <- read.table(obj2.outfile.bgen, header = T, as.is = T)
#  expect_equal(obj2.gds[,-(2:3)], obj2.bgen[,-2])
  unlink(c(obj2.outfile.gds, obj2.outfile.bgen))

  obj2.glmm.score.outfile.gds <- tempfile()
  glmm.score(obj=obj2, infile=gdsfile, outfile=obj2.glmm.score.outfile.gds)
  obj2.glmm.score.gds <- read.table(obj2.glmm.score.outfile.gds,header=T,as.is=T)
  expect_equal(obj2.gds$PVAL.MARGINAL, obj2.glmm.score.gds$PVAL)
  obj2.glmm.score.outfile.bgen <- tempfile()
  glmm.score(obj=obj2, infile=bgenfile, BGEN.samplefile = samplefile, outfile=obj2.glmm.score.outfile.bgen)
  obj2.glmm.score.bgen <- read.table(obj2.glmm.score.outfile.bgen,header=T,as.is=T)
  expect_equal(signif(obj2.gds$PVAL.MARGINAL), signif(obj2.glmm.score.bgen$PVAL))
  unlink(c(obj2.glmm.score.outfile.gds, obj2.glmm.score.outfile.bgen))

  ### multi-thread
  skip_on_cran()
  obj1.outfile.gds.tmp <- tempfile()
  glmm.gei(null.obj=obj1, interaction="sex", geno.file=gdsfile, outfile=obj1.outfile.gds.tmp, ncores=2)
  obj1.gds.tmp <- read.table(obj1.outfile.gds.tmp, header = T, as.is = T)
  expect_equal(obj1.gds, obj1.gds.tmp)
  obj1.outfile.bgen.tmp <- tempfile()
  glmm.gei(null.obj=obj1, interaction="sex", geno.file=bgenfile, bgen.samplefile = samplefile, outfile=obj1.outfile.bgen.tmp, ncores=2)
  obj1.bgen.tmp <- read.table(obj1.outfile.bgen.tmp, header = T, as.is = T)
  expect_equal(obj1.bgen, obj1.bgen.tmp)
  unlink(c(obj1.outfile.gds.tmp, obj1.outfile.bgen.tmp))

  obj2.outfile.gds.tmp <- tempfile()
  glmm.gei(null.obj=obj2, interaction="sex", geno.file=gdsfile, outfile=obj2.outfile.gds.tmp, ncores=2)
  obj2.gds.tmp <- read.table(obj2.outfile.gds.tmp, header = T, as.is = T)
  expect_equal(obj2.gds, obj2.gds.tmp)
  obj2.outfile.bgen.tmp <- tempfile()
  glmm.gei(null.obj=obj2, interaction="sex", geno.file=bgenfile, bgen.samplefile = samplefile, outfile=obj2.outfile.bgen.tmp, ncores=2)
  obj2.bgen.tmp <- read.table(obj2.outfile.bgen.tmp, header = T, as.is = T)
  expect_equal(obj2.bgen, obj2.bgen.tmp)
  unlink(c(obj2.outfile.gds.tmp, obj2.outfile.bgen.tmp))
  
  ### re-order id
  idx <- sample(nrow(pheno))
  pheno <- pheno[idx, ]
  obj1 <- glmmkin(trait ~ age + sex, data = pheno, kins = kins, id = "id", family = gaussian(link = "identity"))
  obj1.outfile.gds <- tempfile()
  glmm.gei(null.obj=obj1, interaction="sex", geno.file=gdsfile, outfile=obj1.outfile.gds)
  tmpout <- read.table(obj1.outfile.gds, header = T, as.is = T)
  expect_equal(obj1.gds, tmpout)
  
  obj1.outfile.bgen <- tempfile()
  glmm.gei(null.obj=obj1, interaction="sex", geno.file=bgenfile, bgen.samplefile = samplefile, outfile=obj1.outfile.bgen)
  tmpout <- read.table(obj1.outfile.bgen, header = T, as.is = T)
#  expect_equal(obj1.bgen, tmpout)
  unlink(c(obj1.outfile.gds, obj1.outfile.bgen))

  obj2 <- glmmkin(trait ~ age + sex, data = pheno, kins = NULL, id = "id", family = gaussian(link = "identity"))
  obj2.outfile.gds <- tempfile()
  glmm.gei(null.obj=obj2, interaction="sex", geno.file=gdsfile, outfile=obj2.outfile.gds)
  
  tmpout <- read.table(obj2.outfile.gds, header = T, as.is = T)
  expect_equal(obj2.gds, tmpout)
  
  obj2.outfile.bgen <- tempfile()
  glmm.gei(null.obj=obj2, interaction="sex", geno.file=bgenfile, bgen.samplefile = samplefile, outfile=obj2.outfile.bgen)
  tmpout <- read.table(obj2.outfile.bgen, header = T, as.is = T)
#  expect_equal(obj2.bgen, tmpout)
  unlink(c(obj2.outfile.gds, obj2.outfile.bgen))
  
  ### re-order id
  idx <- sample(nrow(kins))
  kins <- kins[idx, idx]
  obj1 <- glmmkin(trait ~ age + sex, data = pheno, kins = kins, id = "id", family = gaussian(link = "identity"))
  obj1.outfile.gds <- tempfile()
  glmm.gei(null.obj=obj1, interaction="sex", geno.file=gdsfile, outfile=obj1.outfile.gds)
  tmpout <- read.table(obj1.outfile.gds, header = T, as.is = T)
  expect_equal(obj1.gds, tmpout)
  
  obj1.outfile.bgen <- tempfile()
  glmm.gei(null.obj=obj1, interaction="sex", geno.file=bgenfile, bgen.samplefile = samplefile, outfile=obj1.outfile.bgen)
  tmpout <- read.table(obj1.outfile.bgen, header = T, as.is = T)
#  expect_equal(obj1.bgen, tmpout)
  unlink(c(obj1.outfile.gds, obj1.outfile.bgen))

  obj2 <- glmmkin(trait ~ age + sex, data = pheno, kins = NULL, id = "id", family = gaussian(link = "identity"))
  obj2.outfile.gds <- tempfile()
  glmm.gei(null.obj=obj2, interaction="sex", geno.file=gdsfile, outfile=obj2.outfile.gds)
  tmpout <- read.table(obj2.outfile.gds, header = T, as.is = T)
  expect_equal(obj2.gds, tmpout)
  
  obj2.outfile.bgen <- tempfile()
  glmm.gei(null.obj=obj2, interaction="sex", geno.file=bgenfile, bgen.samplefile = samplefile, outfile=obj2.outfile.bgen)
  tmpout <- read.table(obj2.outfile.bgen, header = T, as.is = T)
#  expect_equal(obj2.bgen, tmpout)
  unlink(c(obj2.outfile.gds, obj2.outfile.bgen))
})

test_that("cross-sectional binomial", {
  skip_on_cran()
  
  gdsfile  <- system.file("extdata", "geno.gds",  package = "MAGEE")
  bgenfile <- system.file("extdata", "geno.bgen", package = "MAGEE")
  samplefile <- system.file("extdata", "geno.sample", package = "MAGEE")
  group.file <- system.file("extdata", "SetID.singlevariant.txt", package = "MAGEE")
  data(example)
  set.seed(123)
  pheno <- rbind(example$pheno, example$pheno[1:100, ])
  pheno$id <- 1:500
  pheno$disease[sample(1:500,20)] <- NA
  pheno$age[sample(1:500,20)] <- NA
  pheno$sex[sample(1:500,20)] <- NA
  pheno <- pheno[sample(1:500,450), ]
  pheno <- pheno[pheno$id <= 400, ]
  kins <- example$GRM
  
  ### single thread with kins
  obj1 <- glmmkin(disease ~ age + sex, data = pheno, kins = kins, id = "id", family = binomial(link = "logit"))
  out1 <- MAGEE(null.obj=obj1, interaction="sex", geno.file=gdsfile, group.file=group.file, tests = c("JV", "JF", "JD"), use.minor.allele = T, auto.flip = T)
  obj1.outfile.gds <- tempfile()
  glmm.gei(null.obj=obj1, interaction="sex", geno.file=gdsfile, outfile=obj1.outfile.gds)
  obj1.gds <- read.table(obj1.outfile.gds, header = T, as.is = T)
  expect_equal(signif(range(obj1.gds$PVAL.MARGINAL)), signif(c(0.008599802, 0.993475648)))
  expect_equal(signif(range(obj1.gds$PVAL.INT)), signif(c(0.01813711, 0.99990766)))
  expect_equal(signif(range(obj1.gds$PVAL.JOINT)), signif(c(0.007484391, 0.998834862)))
  expect_equal(signif(out1$MV.pval,5), signif(obj1.gds$PVAL.MARGINAL,5))
  expect_equal(signif(out1$MF.pval), signif(obj1.gds$PVAL.MARGINAL))
  expect_equal(signif(out1$IV.pval,5), signif(obj1.gds$PVAL.INT,5))
  expect_equal(signif(out1$IF.pval), signif(obj1.gds$PVAL.INT))
  expect_equal(signif(pchisq(qchisq(out1$MV.pval,2,lower=F)+qchisq(out1$IV.pval,2,lower=F),4,lower=F)), signif(out1$JV.pval))
  expect_equal(signif(out1$JV.pval,4), signif(out1$JF.pval,4))
  expect_equal(signif(out1$JV.pval,4), signif(out1$JD.pval,4))
  expect_equal(signif(pchisq(qchisq(obj1.gds$PVAL.MARGINAL,1,lower=F)+qchisq(obj1.gds$PVAL.INT,1,lower=F),2,lower=F)), signif(obj1.gds$PVAL.JOINT))

  obj1.outfile.bgen <- tempfile()
  glmm.gei(null.obj=obj1, interaction="sex", geno.file=bgenfile, bgen.samplefile = samplefile, outfile=obj1.outfile.bgen)
  obj1.bgen <- read.table(obj1.outfile.bgen, header = T, as.is = T)
#  expect_equal(obj1.gds[,-(2:3)], obj1.bgen[,-2])
  unlink(c(obj1.outfile.gds, obj1.outfile.bgen))

  obj1.glmm.score.outfile.gds <- tempfile()
  glmm.score(obj=obj1, infile=gdsfile, outfile=obj1.glmm.score.outfile.gds)
  obj1.glmm.score.gds <- read.table(obj1.glmm.score.outfile.gds,header=T,as.is=T)
  expect_equal(obj1.gds$PVAL.MARGINAL, obj1.glmm.score.gds$PVAL)
  obj1.glmm.score.outfile.bgen <- tempfile()
  glmm.score(obj=obj1, infile=bgenfile, BGEN.samplefile = samplefile, outfile=obj1.glmm.score.outfile.bgen)
  obj1.glmm.score.bgen <- read.table(obj1.glmm.score.outfile.bgen,header=T,as.is=T)
  expect_equal(signif(obj1.gds$PVAL.MARGINAL), signif(obj1.glmm.score.bgen$PVAL))
  unlink(c(obj1.glmm.score.outfile.gds, obj1.glmm.score.outfile.bgen))

  ### single thread without kins
  obj2 <- glmmkin(disease ~ age + sex, data = pheno, kins = NULL, id = "id", family = binomial(link = "logit"))
  out2 <- MAGEE(null.obj=obj2, interaction="sex", geno.file=gdsfile, group.file=group.file, tests = c("JV", "JF", "JD"), use.minor.allele = T, auto.flip = T)
  obj2.outfile.gds <- tempfile()
  glmm.gei(null.obj=obj2, interaction="sex", geno.file=gdsfile, outfile=obj2.outfile.gds)
  obj2.gds <- read.table(obj2.outfile.gds, header = T, as.is = T)
  expect_equal(signif(range(obj2.gds$PVAL.MARGINAL)), signif(c(0.003470366, 0.973400784)))
  expect_equal(signif(range(obj2.gds$PVAL.INT)), signif(c(0.01136652, 0.98224264)))
  expect_equal(signif(range(obj2.gds$PVAL.JOINT)), signif(c(0.003101226, 0.993232042)))
  expect_equal(signif(out2$MV.pval,4), signif(obj2.gds$PVAL.MARGINAL,4))
  expect_equal(signif(out2$MF.pval,4), signif(obj2.gds$PVAL.MARGINAL,4))
  expect_equal(signif(out2$IV.pval,3), signif(obj2.gds$PVAL.INT,3))
  expect_equal(signif(out2$IF.pval,3), signif(obj2.gds$PVAL.INT,3))
  expect_equal(signif(pchisq(qchisq(out2$MV.pval,2,lower=F)+qchisq(out2$IV.pval,2,lower=F),4,lower=F)), signif(out2$JV.pval))
  expect_equal(signif(out2$JV.pval,3), signif(out2$JF.pval,3))
  expect_equal(signif(out2$JV.pval,3), signif(out2$JD.pval,3))
  expect_equal(signif(pchisq(qchisq(obj2.gds$PVAL.MARGINAL,1,lower=F)+qchisq(obj2.gds$PVAL.INT,1,lower=F),2,lower=F)), signif(obj2.gds$PVAL.JOINT))

  obj2.outfile.bgen <- tempfile()
  glmm.gei(null.obj=obj2, interaction="sex", geno.file=bgenfile, bgen.samplefile = samplefile, outfile=obj2.outfile.bgen)
  obj2.bgen <- read.table(obj2.outfile.bgen, header = T, as.is = T)
#  expect_equal(obj2.gds[,-(2:3)], obj2.bgen[,-2])
  unlink(c(obj2.outfile.gds, obj2.outfile.bgen))
  
  obj2.glmm.score.outfile.gds <- tempfile()
  glmm.score(obj=obj2, infile=gdsfile, outfile=obj2.glmm.score.outfile.gds)
  obj2.glmm.score.gds <- read.table(obj2.glmm.score.outfile.gds,header=T,as.is=T)
  expect_equal(obj2.gds$PVAL.MARGINAL, obj2.glmm.score.gds$PVAL)
  obj2.glmm.score.outfile.bgen <- tempfile()
  glmm.score(obj=obj2, infile=bgenfile, BGEN.samplefile = samplefile, outfile=obj2.glmm.score.outfile.bgen)
  obj2.glmm.score.bgen <- read.table(obj2.glmm.score.outfile.bgen,header=T,as.is=T)
  expect_equal(signif(obj2.gds$PVAL.MARGINAL), signif(obj2.glmm.score.bgen$PVAL))
  unlink(c(obj2.glmm.score.outfile.gds, obj2.glmm.score.outfile.bgen))

  ### multi-thread
  obj1.outfile.gds.tmp <- tempfile()
  glmm.gei(null.obj=obj1, interaction="sex", geno.file=gdsfile, outfile=obj1.outfile.gds.tmp, ncores=2)
  obj1.gds.tmp <- read.table(obj1.outfile.gds.tmp, header = T, as.is = T)
  expect_equal(obj1.gds, obj1.gds.tmp)
  obj1.outfile.bgen.tmp <- tempfile()
  glmm.gei(null.obj=obj1, interaction="sex", geno.file=bgenfile, bgen.samplefile = samplefile, outfile=obj1.outfile.bgen.tmp, ncores=2)
  obj1.bgen.tmp <- read.table(obj1.outfile.bgen.tmp, header = T, as.is = T)
  expect_equal(obj1.bgen, obj1.bgen.tmp)
  unlink(c(obj1.outfile.gds.tmp, obj1.outfile.bgen.tmp))

  obj2.outfile.gds.tmp <- tempfile()
  glmm.gei(null.obj=obj2, interaction="sex", geno.file=gdsfile, outfile=obj2.outfile.gds.tmp, ncores=2)
  obj2.gds.tmp <- read.table(obj2.outfile.gds.tmp, header = T, as.is = T)
  expect_equal(obj2.gds, obj2.gds.tmp)
  obj2.outfile.bgen.tmp <- tempfile()
  glmm.gei(null.obj=obj2, interaction="sex", geno.file=bgenfile, bgen.samplefile = samplefile, outfile=obj2.outfile.bgen.tmp, ncores=2)
  obj2.bgen.tmp <- read.table(obj2.outfile.bgen.tmp, header = T, as.is = T)
  expect_equal(obj2.bgen, obj2.bgen.tmp)
  unlink(c(obj2.outfile.gds.tmp, obj2.outfile.bgen.tmp))
  
  ### re-order id
  idx <- sample(nrow(pheno))
  pheno <- pheno[idx, ]
  obj1 <- glmmkin(disease ~ age + sex, data = pheno, kins = kins, id = "id", family = binomial(link = "logit"))
  obj1.outfile.gds <- tempfile()
  glmm.gei(null.obj=obj1, interaction="sex", geno.file=gdsfile, outfile=obj1.outfile.gds)
  tmpout <- read.table(obj1.outfile.gds, header = T, as.is = T)
  expect_equal(obj1.gds, tmpout)
  
  obj1.outfile.bgen <- tempfile()
  glmm.gei(null.obj=obj1, interaction="sex", geno.file=bgenfile, bgen.samplefile = samplefile, outfile=obj1.outfile.bgen)
  tmpout <- read.table(obj1.outfile.bgen, header = T, as.is = T)
#  expect_equal(obj1.bgen, tmpout)
  unlink(c(obj1.outfile.gds, obj1.outfile.bgen))

  obj2 <- glmmkin(disease ~ age + sex, data = pheno, kins = NULL, id = "id", family = binomial(link = "logit"))
  obj2.outfile.gds <- tempfile()
  glmm.gei(null.obj=obj2, interaction="sex", geno.file=gdsfile, outfile=obj2.outfile.gds)
  tmpout <- read.table(obj2.outfile.gds, header = T, as.is = T)
  expect_equal(obj2.gds, tmpout)
  
  obj2.outfile.bgen <- tempfile()
  glmm.gei(null.obj=obj2, interaction="sex", geno.file=bgenfile, bgen.samplefile = samplefile, outfile=obj2.outfile.bgen)
  tmpout <- read.table(obj2.outfile.bgen, header = T, as.is = T)
#  expect_equal(obj2.bgen, tmpout)
  unlink(c(obj2.outfile.gds, obj2.outfile.bgen))
  
  ### re-order id
  idx <- sample(nrow(kins))
  kins <- kins[idx, idx]
  obj1 <- glmmkin(disease ~ age + sex, data = pheno, kins = kins, id = "id", family = binomial(link = "logit"))
  obj1.outfile.gds <- tempfile()
  glmm.gei(null.obj=obj1, interaction="sex", geno.file=gdsfile, outfile=obj1.outfile.gds)
  tmpout <- read.table(obj1.outfile.gds, header = T, as.is = T)
  expect_equal(obj1.gds, tmpout)
  
  obj1.outfile.bgen <- tempfile()
  glmm.gei(null.obj=obj1, interaction="sex", geno.file=bgenfile, bgen.samplefile = samplefile, outfile=obj1.outfile.bgen)
  tmpout <- read.table(obj1.outfile.bgen, header = T, as.is = T)
#  expect_equal(obj1.bgen, tmpout)
  unlink(c(obj1.outfile.gds, obj1.outfile.bgen))

  obj2 <- glmmkin(disease ~ age + sex, data = pheno, kins = NULL, id = "id", family = binomial(link = "logit"))
  obj2.outfile.gds <- tempfile()
  glmm.gei(null.obj=obj2, interaction="sex", geno.file=gdsfile, outfile=obj2.outfile.gds)
  tmpout <- read.table(obj2.outfile.gds, header = T, as.is = T)
  expect_equal(obj2.gds, tmpout)
  
  obj2.outfile.bgen <- tempfile()
  glmm.gei(null.obj=obj2, interaction="sex", geno.file=bgenfile, bgen.samplefile = samplefile, outfile=obj2.outfile.bgen)
  tmpout <- read.table(obj2.outfile.bgen, header = T, as.is = T)
#  expect_equal(obj2.bgen, tmpout)
  unlink(c(obj2.outfile.gds, obj2.outfile.bgen))
})

### longitudinal
test_that("longitudinal random time trend gaussian", {
  skip_on_cran()
  
  gdsfile  <- system.file("extdata", "geno.gds",  package = "MAGEE")
  bgenfile <- system.file("extdata", "geno.bgen", package = "MAGEE")
  samplefile <- system.file("extdata", "geno.sample", package = "MAGEE")
  group.file <- system.file("extdata", "SetID.singlevariant.txt", package = "MAGEE")
  data(example)
  set.seed(123)
  pheno <- example$pheno2
  pheno$y.trend[sample(1:2000,20)] <- NA
  pheno$sex[sample(1:2000,20)] <- NA
  pheno$time[sample(1:2000,20)] <- NA
  pheno <- pheno[sample(1:2000,1950), ]
  kins <- example$GRM
  ### single thread with kins
  obj1 <- glmmkin(y.trend ~ sex + time, data = pheno, kins = kins, id = "id",random.slope = "time", family = gaussian(link = "identity"))
  out1 <- MAGEE(null.obj=obj1, interaction="sex", geno.file=gdsfile, group.file=group.file, tests = c("JV", "JF", "JD"), use.minor.allele = T, auto.flip = T)
  obj1.outfile.gds <- tempfile()
  glmm.gei(null.obj=obj1, interaction="sex", geno.file=gdsfile, outfile=obj1.outfile.gds)
  obj1.gds <- read.table(obj1.outfile.gds, header = T, as.is = T)
  expect_equal(signif(range(obj1.gds$PVAL.MARGINAL)), signif(c(0.01751495, 0.99543498)))
  expect_equal(signif(range(obj1.gds$PVAL.INT)), signif(c(0.008326619, 0.962298286)))
  expect_equal(signif(range(obj1.gds$PVAL.JOINT)), signif(c(0.01331683, 0.99682331)))
  expect_equal(signif(out1$MV.pval,5), signif(obj1.gds$PVAL.MARGINAL,5))
  expect_equal(signif(out1$MF.pval), signif(obj1.gds$PVAL.MARGINAL))
  expect_equal(signif(out1$IV.pval,4), signif(obj1.gds$PVAL.INT,4))
  expect_equal(signif(out1$IF.pval), signif(obj1.gds$PVAL.INT))
  expect_equal(signif(pchisq(qchisq(out1$MV.pval,2,lower=F)+qchisq(out1$IV.pval,2,lower=F),4,lower=F)), signif(out1$JV.pval))
  expect_equal(signif(out1$JV.pval,4), signif(out1$JF.pval,4))
  expect_equal(signif(out1$JV.pval,4), signif(out1$JD.pval,4))
  expect_equal(signif(pchisq(qchisq(obj1.gds$PVAL.MARGINAL,1,lower=F)+qchisq(obj1.gds$PVAL.INT,1,lower=F),2,lower=F)), signif(obj1.gds$PVAL.JOINT))

  obj1.outfile.bgen <- tempfile()
  glmm.gei(null.obj=obj1, interaction="sex", geno.file=bgenfile, bgen.samplefile = samplefile, outfile=obj1.outfile.bgen)
  obj1.bgen <- read.table(obj1.outfile.bgen, header = T, as.is = T)
#  expect_equal(obj1.gds[,-(2:3)], obj1.bgen[,-2])
  unlink(c(obj1.outfile.gds, obj1.outfile.bgen))

  obj1.glmm.score.outfile.gds <- tempfile()
  glmm.score(obj=obj1, infile=gdsfile, outfile=obj1.glmm.score.outfile.gds)
  obj1.glmm.score.gds <- read.table(obj1.glmm.score.outfile.gds,header=T,as.is=T)
  expect_equal(obj1.gds$PVAL.MARGINAL, obj1.glmm.score.gds$PVAL)
  obj1.glmm.score.outfile.bgen <- tempfile()
  glmm.score(obj=obj1, infile=bgenfile, BGEN.samplefile = samplefile, outfile=obj1.glmm.score.outfile.bgen)
  obj1.glmm.score.bgen <- read.table(obj1.glmm.score.outfile.bgen,header=T,as.is=T)
  expect_equal(signif(obj1.gds$PVAL.MARGINAL), signif(obj1.glmm.score.bgen$PVAL))
  unlink(c(obj1.glmm.score.outfile.gds, obj1.glmm.score.outfile.bgen))

  out1.time <- MAGEE(null.obj=obj1, interaction="time", geno.file=gdsfile, group.file=group.file, tests = c("JV", "JF", "JD"), use.minor.allele = T, auto.flip = T)
  obj1.outfile.gds <- tempfile()
  glmm.gei(null.obj=obj1, interaction="time", geno.file=gdsfile, outfile=obj1.outfile.gds)
  obj1.time.gds <- read.table(obj1.outfile.gds, header = T, as.is = T)
  expect_equal(obj1.gds$PVAL.MARGINAL, obj1.time.gds$PVAL.MARGINAL)
  expect_equal(signif(range(obj1.time.gds$PVAL.INT)), signif(c(0.01411984, 0.97927828)))
  expect_equal(signif(range(obj1.time.gds$PVAL.JOINT)), signif(c(0.02898685, 0.98741654)))
  expect_equal(out1$MV.pval, out1.time$MV.pval)
  expect_equal(out1$MF.pval, out1.time$MF.pval)
  expect_equal(signif(out1.time$IV.pval,3), signif(obj1.time.gds$PVAL.INT,3))
  expect_equal(signif(out1.time$IF.pval), signif(obj1.time.gds$PVAL.INT))
  expect_equal(signif(pchisq(qchisq(out1.time$MV.pval,2,lower=F)+qchisq(out1.time$IV.pval,2,lower=F),4,lower=F)), signif(out1.time$JV.pval))
  expect_equal(signif(out1.time$JV.pval,3), signif(out1.time$JF.pval,3))
  expect_equal(signif(out1.time$JV.pval,3), signif(out1.time$JD.pval,3))
  expect_equal(signif(pchisq(qchisq(obj1.time.gds$PVAL.MARGINAL,1,lower=F)+qchisq(obj1.time.gds$PVAL.INT,1,lower=F),2,lower=F)), signif(obj1.time.gds$PVAL.JOINT))

  obj1.outfile.bgen <- tempfile()
  glmm.gei(null.obj=obj1, interaction="time", geno.file=bgenfile, bgen.samplefile = samplefile, outfile=obj1.outfile.bgen)
  obj1.time.bgen <- read.table(obj1.outfile.bgen, header = T, as.is = T)
#  expect_equal(obj1.time.gds[,-(2:3)], obj1.time.bgen[,-2])
  unlink(c(obj1.outfile.gds, obj1.outfile.bgen))

  ### single thread without kins
  obj2 <- glmmkin(y.trend ~ sex + time, data = pheno, kins = NULL, id = "id",random.slope = "time", family = gaussian(link = "identity"))
  out2 <- MAGEE(null.obj=obj2, interaction="sex", geno.file=gdsfile, group.file=group.file, tests = c("JV", "JF", "JD"), use.minor.allele = T, auto.flip = T)
  obj2.outfile.gds <- tempfile()
  glmm.gei(null.obj=obj2, interaction="sex", geno.file=gdsfile, outfile=obj2.outfile.gds)
  obj2.gds <- read.table(obj2.outfile.gds, header = T, as.is = T)
  expect_equal(signif(range(obj2.gds$PVAL.MARGINAL)), signif(c(0.008752958, 0.999386091)))
  expect_equal(signif(range(obj2.gds$PVAL.INT)), signif(c(0.01365968, 0.99280013)))
  expect_equal(signif(range(obj2.gds$PVAL.JOINT)), signif(c(0.01236586, 0.99750905)))
  expect_equal(signif(out2$MV.pval,4), signif(obj2.gds$PVAL.MARGINAL,4))
  expect_equal(signif(out2$MF.pval), signif(obj2.gds$PVAL.MARGINAL))
  expect_equal(signif(out2$IV.pval,4), signif(obj2.gds$PVAL.INT,4))
  expect_equal(signif(out2$IF.pval), signif(obj2.gds$PVAL.INT))
  expect_equal(signif(pchisq(qchisq(out2$MV.pval,2,lower=F)+qchisq(out2$IV.pval,2,lower=F),4,lower=F)), signif(out2$JV.pval))
  expect_equal(signif(out2$JV.pval,4), signif(out2$JF.pval,4))
  expect_equal(signif(out2$JV.pval,4), signif(out2$JD.pval,4))
  expect_equal(signif(pchisq(qchisq(obj2.gds$PVAL.MARGINAL,1,lower=F)+qchisq(obj2.gds$PVAL.INT,1,lower=F),2,lower=F)), signif(obj2.gds$PVAL.JOINT))

  obj2.outfile.bgen <- tempfile()
  glmm.gei(null.obj=obj2, interaction="sex", geno.file=bgenfile, bgen.samplefile = samplefile, outfile=obj2.outfile.bgen)
  obj2.bgen <- read.table(obj2.outfile.bgen, header = T, as.is = T)
#  expect_equal(obj2.gds[,-(2:3)], obj2.bgen[,-2])
  unlink(c(obj2.outfile.gds, obj2.outfile.bgen))

  obj2.glmm.score.outfile.gds <- tempfile()
  glmm.score(obj=obj2, infile=gdsfile, outfile=obj2.glmm.score.outfile.gds)
  obj2.glmm.score.gds <- read.table(obj2.glmm.score.outfile.gds,header=T,as.is=T)
  expect_equal(obj2.gds$PVAL.MARGINAL, obj2.glmm.score.gds$PVAL)
  obj2.glmm.score.outfile.bgen <- tempfile()
  glmm.score(obj=obj2, infile=bgenfile, BGEN.samplefile = samplefile, outfile=obj2.glmm.score.outfile.bgen)
  obj2.glmm.score.bgen <- read.table(obj2.glmm.score.outfile.bgen,header=T,as.is=T)
  expect_equal(signif(obj2.gds$PVAL.MARGINAL), signif(obj2.glmm.score.bgen$PVAL))
  unlink(c(obj2.glmm.score.outfile.gds, obj2.glmm.score.outfile.bgen))

  out2.time <- MAGEE(null.obj=obj2, interaction="time", geno.file=gdsfile, group.file=group.file, tests = c("JV", "JF", "JD"), use.minor.allele = T, auto.flip = T)
  obj2.outfile.gds <- tempfile()
  glmm.gei(null.obj=obj2, interaction="time", geno.file=gdsfile, outfile=obj2.outfile.gds)
  obj2.time.gds <- read.table(obj2.outfile.gds, header = T, as.is = T)
  expect_equal(obj2.gds$PVAL.MARGINAL, obj2.time.gds$PVAL.MARGINAL)
  expect_equal(signif(range(obj2.time.gds$PVAL.INT)), signif(c(0.006739873, 0.995617773)))
  expect_equal(signif(range(obj2.time.gds$PVAL.JOINT)), signif(c(0.01342866, 0.99567692)))
  expect_equal(out2$MV.pval, out2.time$MV.pval)
  expect_equal(out2$MF.pval, out2.time$MF.pval)
  expect_equal(signif(out2.time$IV.pval,5), signif(obj2.time.gds$PVAL.INT,5))
  expect_equal(signif(out2.time$IF.pval), signif(obj2.time.gds$PVAL.INT))
  expect_equal(signif(pchisq(qchisq(out2.time$MV.pval,2,lower=F)+qchisq(out2.time$IV.pval,2,lower=F),4,lower=F)), signif(out2.time$JV.pval))
  expect_equal(signif(out2.time$JV.pval,3), signif(out2.time$JF.pval,3))
  expect_equal(signif(out2.time$JV.pval,3), signif(out2.time$JD.pval,3))
  expect_equal(signif(pchisq(qchisq(obj2.time.gds$PVAL.MARGINAL,1,lower=F)+qchisq(obj2.time.gds$PVAL.INT,1,lower=F),2,lower=F)), signif(obj2.time.gds$PVAL.JOINT))

  obj2.outfile.bgen <- tempfile()
  glmm.gei(null.obj=obj2, interaction="time", geno.file=bgenfile, bgen.samplefile = samplefile, outfile=obj2.outfile.bgen)
  obj2.time.bgen <- read.table(obj2.outfile.bgen, header = T, as.is = T)
#  expect_equal(obj2.time.gds[,-(2:3)], obj2.time.bgen[,-2])
  unlink(c(obj2.outfile.gds, obj2.outfile.bgen))

  ### multi-thread
  obj1.outfile.gds.tmp <- tempfile()
  glmm.gei(null.obj=obj1, interaction="sex", geno.file=gdsfile, outfile=obj1.outfile.gds.tmp, ncores=2)
  obj1.gds.tmp <- read.table(obj1.outfile.gds.tmp, header = T, as.is = T)
  expect_equal(obj1.gds, obj1.gds.tmp)
  obj1.outfile.bgen.tmp <- tempfile()
  glmm.gei(null.obj=obj1, interaction="sex", geno.file=bgenfile, bgen.samplefile = samplefile, outfile=obj1.outfile.bgen.tmp, ncores=2)
  obj1.bgen.tmp <- read.table(obj1.outfile.bgen.tmp, header = T, as.is = T)
#  expect_equal(obj1.bgen, obj1.bgen.tmp)
  unlink(c(obj1.outfile.gds.tmp, obj1.outfile.bgen.tmp))

  obj1.outfile.gds.tmp <- tempfile()
  glmm.gei(null.obj=obj1, interaction="time", geno.file=gdsfile, outfile=obj1.outfile.gds.tmp, ncores=2)
  obj1.time.gds.tmp <- read.table(obj1.outfile.gds.tmp, header = T, as.is = T)
  expect_equal(obj1.time.gds, obj1.time.gds.tmp)
  obj1.outfile.bgen.tmp <- tempfile()
  glmm.gei(null.obj=obj1, interaction="time", geno.file=bgenfile, bgen.samplefile = samplefile, outfile=obj1.outfile.bgen.tmp, ncores=2)
  obj1.time.bgen.tmp <- read.table(obj1.outfile.bgen.tmp, header = T, as.is = T)
#  expect_equal(obj1.time.bgen, obj1.time.bgen.tmp)
  unlink(c(obj1.outfile.gds.tmp, obj1.outfile.bgen.tmp))

  obj2.outfile.gds.tmp <- tempfile()
  glmm.gei(null.obj=obj2, interaction="sex", geno.file=gdsfile, outfile=obj2.outfile.gds.tmp, ncores=2)
  obj2.gds.tmp <- read.table(obj2.outfile.gds.tmp, header = T, as.is = T)
  expect_equal(obj2.gds, obj2.gds.tmp)
  obj2.outfile.bgen.tmp <- tempfile()
  glmm.gei(null.obj=obj2, interaction="sex", geno.file=bgenfile, bgen.samplefile = samplefile, outfile=obj2.outfile.bgen.tmp, ncores=2)
  obj2.bgen.tmp <- read.table(obj2.outfile.bgen.tmp, header = T, as.is = T)
#  expect_equal(obj2.bgen, obj2.bgen.tmp)
  unlink(c(obj2.outfile.gds.tmp, obj2.outfile.bgen.tmp))
  
  obj2.outfile.gds.tmp <- tempfile()
  glmm.gei(null.obj=obj2, interaction="time", geno.file=gdsfile, outfile=obj2.outfile.gds.tmp, ncores=2)
  obj2.time.gds.tmp <- read.table(obj2.outfile.gds.tmp, header = T, as.is = T)
  expect_equal(obj2.time.gds, obj2.time.gds.tmp)
  obj2.outfile.bgen.tmp <- tempfile()
  glmm.gei(null.obj=obj2, interaction="time", geno.file=bgenfile, bgen.samplefile = samplefile, outfile=obj2.outfile.bgen.tmp, ncores=2)
  obj2.time.bgen.tmp <- read.table(obj2.outfile.bgen.tmp, header = T, as.is = T)
#  expect_equal(obj2.time.bgen, obj2.time.bgen.tmp)
  unlink(c(obj2.outfile.gds.tmp, obj2.outfile.bgen.tmp))
  
### re-order id
  idx <- sample(nrow(pheno))
  pheno <- pheno[idx, ]
  obj1 <- glmmkin(y.trend ~ sex + time, data = pheno, kins = kins, id = "id",random.slope = "time", family = gaussian(link = "identity"))
  obj1.outfile.gds <- tempfile()
  glmm.gei(null.obj=obj1, interaction="sex", geno.file=gdsfile, outfile=obj1.outfile.gds)
  tmpout <- read.table(obj1.outfile.gds, header = T, as.is = T)
  expect_equal(obj1.gds, tmpout)
  
  obj1.outfile.bgen <- tempfile()
  glmm.gei(null.obj=obj1, interaction="sex", geno.file=bgenfile, bgen.samplefile = samplefile, outfile=obj1.outfile.bgen)
  tmpout <- read.table(obj1.outfile.bgen, header = T, as.is = T)
#  expect_equal(obj1.bgen, tmpout)
  unlink(c(obj1.outfile.gds, obj1.outfile.bgen))

  obj1.outfile.gds <- tempfile()
  glmm.gei(null.obj=obj1, interaction="time", geno.file=gdsfile, outfile=obj1.outfile.gds)
  tmpout <- read.table(obj1.outfile.gds, header = T, as.is = T)
  expect_equal(obj1.time.gds, tmpout)
  
  obj1.outfile.bgen <- tempfile()
  glmm.gei(null.obj=obj1, interaction="time", geno.file=bgenfile, bgen.samplefile = samplefile, outfile=obj1.outfile.bgen)
  tmpout <- read.table(obj1.outfile.bgen, header = T, as.is = T)
#  expect_equal(obj1.time.bgen, tmpout)
  unlink(c(obj1.outfile.gds, obj1.outfile.bgen))

  obj2 <- glmmkin(y.trend ~ sex + time, data = pheno, kins = NULL, id = "id",random.slope = "time", family = gaussian(link = "identity"))
  obj2.outfile.gds <- tempfile()
  glmm.gei(null.obj=obj2, interaction="sex", geno.file=gdsfile, outfile=obj2.outfile.gds)
  tmpout <- read.table(obj2.outfile.gds, header = T, as.is = T)
  expect_equal(obj2.gds, tmpout)
  
  obj2.outfile.bgen <- tempfile()
  glmm.gei(null.obj=obj2, interaction="sex", geno.file=bgenfile, bgen.samplefile = samplefile, outfile=obj2.outfile.bgen)
  tmpout <- read.table(obj2.outfile.bgen, header = T, as.is = T)
#  expect_equal(obj2.bgen, tmpout)
  unlink(c(obj2.outfile.gds, obj2.outfile.bgen))

  obj2.outfile.gds <- tempfile()
  glmm.gei(null.obj=obj2, interaction="time", geno.file=gdsfile, outfile=obj2.outfile.gds)
  tmpout <- read.table(obj2.outfile.gds, header = T, as.is = T)
  expect_equal(obj2.time.gds, tmpout)
  
  obj2.outfile.bgen <- tempfile()
  glmm.gei(null.obj=obj2, interaction="time", geno.file=bgenfile, bgen.samplefile = samplefile, outfile=obj2.outfile.bgen)
  tmpout <- read.table(obj2.outfile.bgen, header = T, as.is = T)
#  expect_equal(obj2.time.bgen, tmpout)
  unlink(c(obj2.outfile.gds, obj2.outfile.bgen))

  ### re-order id
  idx <- sample(nrow(kins))
  kins <- kins[idx, idx]
  obj1 <- glmmkin(y.trend ~ sex + time, data = pheno, kins = kins, id = "id",random.slope = "time", family = gaussian(link = "identity"))
  obj1.outfile.gds <- tempfile()
  glmm.gei(null.obj=obj1, interaction="sex", geno.file=gdsfile, outfile=obj1.outfile.gds)
  tmpout <- read.table(obj1.outfile.gds, header = T, as.is = T)
  expect_equal(obj1.gds, tmpout)
  
  obj1.outfile.bgen <- tempfile()
  glmm.gei(null.obj=obj1, interaction="sex", geno.file=bgenfile, bgen.samplefile = samplefile, outfile=obj1.outfile.bgen)
  tmpout <- read.table(obj1.outfile.bgen, header = T, as.is = T)
#  expect_equal(obj1.bgen, tmpout)
  unlink(c(obj1.outfile.gds, obj1.outfile.bgen))

  obj1.outfile.gds <- tempfile()
  glmm.gei(null.obj=obj1, interaction="time", geno.file=gdsfile, outfile=obj1.outfile.gds)
  tmpout <- read.table(obj1.outfile.gds, header = T, as.is = T)
  expect_equal(obj1.time.gds, tmpout)
  
  obj1.outfile.bgen <- tempfile()
  glmm.gei(null.obj=obj1, interaction="time", geno.file=bgenfile, bgen.samplefile = samplefile, outfile=obj1.outfile.bgen)
  tmpout <- read.table(obj1.outfile.bgen, header = T, as.is = T)
#  expect_equal(obj1.time.bgen, tmpout)
  unlink(c(obj1.outfile.gds, obj1.outfile.bgen))

  obj2 <- glmmkin(y.trend ~ sex + time, data = pheno, kins = NULL, id = "id",random.slope = "time", family = gaussian(link = "identity"))
  obj2.outfile.gds <- tempfile()
  glmm.gei(null.obj=obj2, interaction="sex", geno.file=gdsfile, outfile=obj2.outfile.gds)
  tmpout <- read.table(obj2.outfile.gds, header = T, as.is = T)
  expect_equal(obj2.gds, tmpout)
  
  obj2.outfile.bgen <- tempfile()
  glmm.gei(null.obj=obj2, interaction="sex", geno.file=bgenfile, bgen.samplefile = samplefile, outfile=obj2.outfile.bgen)
  tmpout <- read.table(obj2.outfile.bgen, header = T, as.is = T)
#  expect_equal(obj2.bgen, tmpout)
  unlink(c(obj2.outfile.gds, obj2.outfile.bgen))

  obj2.outfile.gds <- tempfile()
  glmm.gei(null.obj=obj2, interaction="time", geno.file=gdsfile, outfile=obj2.outfile.gds)
  tmpout <- read.table(obj2.outfile.gds, header = T, as.is = T)
  expect_equal(obj2.time.gds, tmpout)
  
  obj2.outfile.bgen <- tempfile()
  glmm.gei(null.obj=obj2, interaction="time", geno.file=bgenfile, bgen.samplefile = samplefile, outfile=obj2.outfile.bgen)
  tmpout <- read.table(obj2.outfile.bgen, header = T, as.is = T)
#  expect_equal(obj2.time.bgen, tmpout)
  unlink(c(obj2.outfile.gds, obj2.outfile.bgen))
})
